#include <Arduino.h>
#include "I2Cdev.h"
#include "MPU6050.h"

#define motor_IN1 3
#define motor_IN2 4
#define motor_IN3 5
#define motor_IN4 6

// 가속도 센서 관련
MPU6050 accelgyro;

// Kalman 필터 관련
class KalmanFilter {
public:
    KalmanFilter() {
        // 변수 초기화
        x_hat = 0;
        P = 1;
        Q = 0.0001;
        R = 0.01;
    }

    float updateEstimate(float z) {
        // 예측 업데이트
        x_hat_minus = x_hat; //필터링 신호 예측 = 사전 예측값
        P_minus = P + Q; // 2차 예측 오차 = 예측 오차 + 프로세스 노이즈

        // 측정 업데이트
        K = P_minus / (P_minus + R); // 칼만 센서 최종값 = 2차 예측 오차 / (2차 예측 오차 + 측정 노이즈)
        x_hat = x_hat_minus + K * (z - x_hat_minus); // 필터링 신호 예측 = 사전 예측값 + 칼만 센서 최종값 * (z축 - 사전 예측값)
        P = (1 - K) * P_minus; // 예측 오차 = (1 - 칼만 센서 최종값) * 2차 예측 오차

        return x_hat; // 필터링된 값 반환
    }

private:
    float x_hat;    // 필터링 신호 예측
    float P;        // 예측 오차
    float Q;        // 프로세스 노이즈 
    float R;        // 측정 노이즈
    float x_hat_minus; // 사전 예측값
    float P_minus;  // 2차 예측 오차
    float K;        // 칼만 센서 최종값
};

// 가속도 센서 값을 저장할 배열
const int numSamples = 10;
int16_t axSamples[numSamples];
int16_t aySamples[numSamples];
int16_t azSamples[numSamples];

// 자이로 센서 값을 저장할 배열
int16_t gxSamples[numSamples];
int16_t gySamples[numSamples];
int16_t gzSamples[numSamples];

// 이동 평균값 변수
int16_t axAvg, ayAvg, azAvg;
int16_t gxAvg, gyAvg, gzAvg;

void setup() {
    Serial.begin(9600);
    Wire.begin();
    accelgyro.initialize();
    pinMode(motor_IN1, OUTPUT);
    pinMode(motor_IN2, OUTPUT);
    pinMode(motor_IN3, OUTPUT);
    pinMode(motor_IN4, OUTPUT);
}

void loop() {
    // 모든 가속도 센서 값을 읽어와서 샘플 배열에 저장
    for (int i = 0; i < numSamples; i++) {
        int16_t ax, ay, az;
        accelgyro.getAcceleration(&ax, &ay, &az);
        axSamples[i] = ax;
        aySamples[i] = ay;
        azSamples[i] = az;

        int16_t gx, gy, gz;
        accelgyro.getRotation(&gx, &gy, &gz);
        gxSamples[i] = gx;
        gySamples[i] = gy;
        gzSamples[i] = gz;

        delay(1); // 필요한 경우 샘플링 간격을 조절
        
    }

    // 칼만 필터를 사용하여 각 축의 가속도 값을 필터링
    float filteredAx = applyKalmanFilter(axSamples, numSamples);
    float filteredAy = applyKalmanFilter(aySamples, numSamples);
    float filteredAz = applyKalmanFilter(azSamples, numSamples);
    
    // 칼만 필터를 사용하여 각 축의 자이로 값을 필터링
    float filteredGx = applyKalmanFilter(gxSamples, numSamples);
    float filteredGy = applyKalmanFilter(gySamples, numSamples);
    float filteredGz = applyKalmanFilter(gzSamples, numSamples);
    
    // 필터링된 가속도 값을 시리얼 모니터에 출력 (한 줄에 출력)
    Serial.print("Filtered Ax: ");
    Serial.print(filteredAx);
    Serial.print("   Filtered Ay: ");
    Serial.print(filteredAy);
    Serial.print("   Filtered Az: ");
    Serial.print(filteredAz);

    // 필터링된 자이로 값을 시리얼 모니터에 출력 (한 줄에 출력)
    Serial.print("   Filtered Gx: ");
    Serial.print(filteredGx);
    Serial.print("   Filtered Gy: ");
    Serial.print(filteredGy);
    Serial.print("   Filtered Gz: ");
    Serial.println(filteredGz);
  int ax1 = filteredAx;
  int ay1 = filteredAy;
  int az1 = filteredAz;
  int speed = random(256);
  analogWrite(motor_IN1, speed);
  analogWrite(motor_IN2, 0);
  analogWrite(motor_IN3, speed);
  analogWrite(motor_IN4, 0);
  delay(3000);

  while (ax1 > filteredAx && ay1 > filteredAy && az1 > filteredAz) {
  analogWrite(motor_IN1, 0);
  analogWrite(motor_IN2, 255);
  analogWrite(motor_IN3, 0);
  analogWrite(motor_IN4, 255);
  }
}

float applyKalmanFilter(int16_t* samples, int numSamples) {
    KalmanFilter filter;
    float filteredValue = 0;

    // 각 샘플을 칼만 필터에 적용하여 필터링된 값 계산
    for (int i = 0; i < numSamples; i++) {
        filteredValue = filter.updateEstimate(samples[i]);
    }
    return filteredValue;
}



